#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib-Datasource
   @package: datasource
      @file: db_repository.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from abc import abstractmethod
from datasource.crud_entity import CrudEntity
from datasource.crud_repository import CrudRepository
from datasource.db_configuration import DBConfiguration
from hspylib.core.metaclass.singleton import AbstractSingleton
from retry import retry
from typing import Any, Generic, Iterable, Optional, Tuple, TypeVar

import contextlib

# fmt: off
# Stereotypes
Connection  = TypeVar("Connection", bound=Any)
Cursor      = TypeVar("Cursor", bound=Any)
Session     = TypeVar("Session", bound=Any)
ResultSet   = TypeVar("ResultSet", bound=Iterable)
# fmt: on

# Generics
E = TypeVar("E", bound=CrudEntity)
C = TypeVar("C", bound=DBConfiguration)


class DBRepository(Generic[E, C], CrudRepository[E], metaclass=AbstractSingleton):
    """Interface for generic CRUD operations on a database for a specific type."""

    def __init__(self, config: C):
        super().__init__()
        self._config = config

    def __str__(self):
        return f"{self.logname}/{self.table_name()} -> {self.info}"

    def __repr__(self):
        return str(self)

    @property
    def config(self) -> C:
        return self._config

    @property
    def info(self) -> str:
        return f"{self.username or ''}@{self.hostname or ''}:{self.port or 0}/{self.database or ''}"

    @property
    def hostname(self) -> str:
        return self._config.hostname

    @property
    def port(self) -> int:
        return self._config.port

    @property
    def username(self) -> str:
        return self._config.username

    @property
    def password(self) -> str:
        return self._config.password

    @property
    def database(self) -> str:
        return self._config.database

    @abstractmethod
    @retry(tries=3, delay=2, backoff=3, max_delay=30)
    def _create_session(self) -> Tuple[Connection, Cursor]:
        """Generator to create a database session and return it."""

    @abstractmethod
    @contextlib.contextmanager
    def _session(self) -> Session:
        """Connect and create a database session."""

    @abstractmethod
    def execute(self, sql_statement: str, **kwargs) -> Tuple[int, Optional[ResultSet]]:
        """Execute the specified SQL statement.
        :param sql_statement: the SQL statement. Names placeholders can be used to parametrize the SQL.
        :param kwargs arbitrary arguments to fill SQL placeholders (names must match the placeholder names).
        """

    @abstractmethod
    def table_name(self) -> str:
        """Return the table name (repository name)."""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib-Datasource
   @package: datasource.firebase
      @file: firebase_configuration.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from collections import defaultdict
from hspylib.core.config.app_config import AppConfigs
from hspylib.core.enums.charset import Charset
from hspylib.core.enums.http_code import HttpCode
from hspylib.core.preconditions import check_argument, check_not_none, check_state
from hspylib.core.tools.commons import dirname
from hspylib.core.tools.dict_tools import get_or_default_by_key
from hspylib.modules.fetch.fetch import get
from os.path import basename
from textwrap import dedent
from typing import Optional

import os


class FirebaseConfiguration(AppConfigs):
    """Represents a Firebase datasource configuration"""

    CONFIG_FORMAT = dedent(
        """
    # Your Firebase configuration:
    # --------------------------
    UID={uid}
    PROJECT_ID={project_id}
    EMAIL={email}
    DATABASE={database}
    BASE_URL={base_url}
    """
    )

    REQUIRED_SETTINGS = ["UID", "PROJECT_ID", "EMAIL", "DATABASE"]

    @staticmethod
    def of(resource_dir: str, filename: str, config_dict: dict) -> "FirebaseConfiguration":
        """Create a Firebase config from a dictionary.
        :param resource_dir the directory where the configuration file is located.
        :param filename the base name of the configuration file.
        :param config_dict a dictionary containing all of the configuration properties.
        """

        check_state(
            all(setting in config_dict for setting in FirebaseConfiguration.REQUIRED_SETTINGS),
            f"Invalid configuration file.\nMust contain required settings: {FirebaseConfiguration.REQUIRED_SETTINGS}",
        )
        return FirebaseConfiguration(
            resource_dir,
            filename,
            config_dict["UID"],
            config_dict["PROJECT_ID"],
            config_dict["EMAIL"],
            config_dict["DATABASE"],
        )

    @staticmethod
    def of_file(filename: str, encoding: str = Charset.UTF_8.val) -> "FirebaseConfiguration":
        """Create a Firebase config from a config file."""
        check_argument(os.path.exists(filename), f"Config file does not exist: {filename}")

        with open(filename, encoding=encoding or Charset.UTF_8.val) as f_config:
            cfg = defaultdict()
            for line in f_config:
                line = line.strip()
                if line.startswith("#") or "=" not in line:
                    continue
                key, value = line.split("=", 1)
                if key in FirebaseConfiguration.REQUIRED_SETTINGS:
                    cfg[key.strip()] = value.strip()

            return FirebaseConfiguration.of(dirname(filename), basename(filename), cfg)

    def __init__(
        self,
        resource_dir: str,
        filename: str,
        uid: str,
        project_id: str,
        email: str,
        database: Optional[str] = None,
        profile: Optional[str] = None,
    ):
        super().__init__(resource_dir, filename, profile)
        self._valid = None
        self._base_url = self["datasource.base.url"]
        self._scheme = self["datasource.scheme"] or "https"
        self._hostname = self["datasource.hostname"] or "firebaseio.com"
        self._port = self["datasource.port"] or 443 if self._scheme.endswith("s") else 80
        self._uid = uid or self["firebase.user.uid"]
        self._project_id = project_id or self["firebase.project.id"]
        self._email = email or self["firebase.email"]
        self._database = database or self["firebase.database"]

    def __str__(self) -> str:
        return (
            ""
            if not hasattr(self, "_uid")
            else self.CONFIG_FORMAT.format(
                uid=self._uid or "not-set",
                project_id=self._project_id or "not-set",
                email=self._email or "not-set",
                database=self.database or "not-set",
                base_url=self.base_url or "not-set",
            )
        )

    @property
    def hostname(self) -> str:
        return f"{self.project_id}.{self._hostname}"

    @property
    def scheme(self) -> str:
        return self._scheme

    @property
    def port(self) -> int:
        return self._port

    @property
    def uid(self) -> str:
        return self._uid

    @property
    def project_id(self) -> str:
        return self._project_id

    @property
    def email(self) -> str:
        return self._email

    @property
    def database(self) -> Optional[str]:
        return self._database

    @property
    def base_url(self) -> str:
        return self._base_url or f"{self.scheme}://{self.hostname}:{self.port}/{self.database}"

    def url(self, db_alias: Optional[str] = "") -> str:
        """Return the url for an element at the Firebase webapp."""
        final_alias_url = db_alias.replace(" ", "%20").replace(".", "/")
        return f"{self.base_url}/{final_alias_url}"

    def validate_config(self) -> bool:
        """Validate the current configuration"""
        check_not_none(self._uid, "User UID must be defined")
        check_not_none(self._project_id, "Project ID must be defined")
        check_not_none(self._email, "Email must be defined")
        response = get(f"{self.base_url}.json")
        self._valid = response is not None and response.status_code == HttpCode.OK

        return self._valid

    def update(self, configs: dict) -> None:
        """Update Firebase configurations"""
        self._uid = get_or_default_by_key(configs, "UID", self.uid)
        self._project_id = get_or_default_by_key(configs, "PROJECT_ID", self.uid)
        self._email = get_or_default_by_key(configs, "EMAIL", self.uid)
        self._database = get_or_default_by_key(configs, "DATABASE", self.uid)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib-Datasource
   @package: datasource.mysql
      @file: mysql_repository.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from abc import abstractmethod
from datasource.crud_entity import CrudEntity
from datasource.db_configuration import DBConfiguration
from datasource.db_repository import DBRepository, ResultSet, Session
from datasource.exception.exceptions import DatabaseConnectionError, DatabaseError
from datasource.identity import Identity
from hspylib.core.metaclass.singleton import AbstractSingleton
from hspylib.core.namespace import Namespace
from hspylib.core.tools.text_tools import quote
from pymysql import Connection, Error
from pymysql.cursors import Cursor
from retry import retry
from typing import Generic, List, Optional, Set, Tuple, TypeVar

import contextlib
import logging as log
import pymysql

E = TypeVar("E", bound=CrudEntity)
C = TypeVar("C", bound=DBConfiguration)


class MySqlRepository(Generic[E, C], DBRepository[E, C], metaclass=AbstractSingleton):
    """Implementation of a data access layer for a MySql persistence store."""

    def __init__(self, config: DBConfiguration):
        super().__init__(config)

    @retry(tries=3, delay=2, backoff=3, max_delay=30)
    def _create_session(self) -> Tuple[Connection, Cursor]:
        log.debug("%s Attempt to connect to database: %s", self.logname, str(self))
        conn = pymysql.connect(
            host=self.hostname, user=self.username, port=self.port, password=self.password, database=self.database
        )
        return conn, conn.cursor()

    @contextlib.contextmanager
    def _session(self) -> Session:
        conn, dbs = None, None
        try:
            conn, dbs = self._create_session()
            log.debug("%s Successfully connected to database: %s [ssid=%d]", self.logname, self.info, hash(dbs))
            yield dbs
        except Error as err:
            raise DatabaseConnectionError(f"Unable to open/execute-on database session => {err}") from err
        finally:
            if conn:
                log.debug("%s Closing connection [ssid=%d]", self.logname, hash(dbs))
                conn.commit()
                conn.close()

    def execute(self, sql_statement: str, **kwargs) -> Tuple[int, Optional[ResultSet]]:
        with self._session() as dbs:
            try:
                args = dict(kwargs)
                rows = []
                log.debug(
                    "%s Executing SQL statement [ssid=%s]:\n" "\t|-Arguments: %s\n" "\t|-Statement: %s",
                    self.logname,
                    hash(dbs),
                    args,
                    sql_statement,
                )
                if (rows_affected := dbs.execute(sql_statement, tuple(args.values())) or 0) > 0:
                    list(map(rows.append, dbs))
                return rows_affected, rows
            except (pymysql.err.ProgrammingError, pymysql.err.OperationalError) as err:
                raise DatabaseError(f"Unable to execute statement => {sql_statement}") from err

    def count(self) -> int:
        sql = f"SELECT COUNT(*) FROM {self.table_name()}"
        return int(self.execute(sql)[1][0][0])

    def delete(self, entity: CrudEntity) -> None:
        self.delete_by_id(entity.identity)

    def delete_by_id(self, entity_id: Identity) -> None:
        clauses = [f"{k} = {quote(v)}" for k, v in zip(entity_id.attributes, entity_id.values)]
        sql = f"DELETE FROM {self.table_name()} WHERE " + " AND ".join(clauses)
        self.execute(sql)

    def delete_all(self, entities: List[E]) -> None:
        values, s = [], entities[0]
        list(map(lambda e: values.append(str(e.values)), entities))
        sql = f"DELETE FROM " f"{self.table_name()} WHERE ({s.as_columns()}) IN ({', '.join(values)}) "
        self.execute(sql)

    def save(self, entity: E) -> None:
        columns, ids = entity.as_columns(), set(entity.identity.attributes)
        sql = (
            f"INSERT INTO "
            f"{self.table_name()} ({columns}) VALUES {entity.values} AS new "
            f"ON DUPLICATE KEY UPDATE {entity.as_column_set(prefix='new.', exclude=ids)}"
        )
        self.execute(sql)

    def save_all(self, entities: List[E]) -> None:
        values, sample = [], entities[0]
        columns, ids = sample.as_columns(), set(sample.identity.attributes)
        list(map(lambda e: values.append(str(e.values)), entities))
        sql = (
            f"INSERT INTO "
            f"{self.table_name()} ({columns}) VALUES {', '.join(values)} AS new "
            f"ON DUPLICATE KEY UPDATE {sample.as_column_set(prefix='new.', exclude=ids)}"
        )
        self.execute(sql)

    def find_all(
        self,
        columns: Optional[Set[str]] = None,
        filters: Optional[Namespace] = None,
        order_bys: Optional[List[str]] = None,
        limit: int = 500,
        offset: int = 0,
    ) -> List[E]:
        columns = "*" if not columns else ", ".join(columns)
        clauses = list(filter(None, filters.values)) if filters else None
        orders = list(filter(None, order_bys)) if order_bys else None
        sql = (
            f"SELECT {columns} FROM {self.table_name()} "
            f"{('WHERE ' + ' AND '.join(clauses)) if clauses else ''} "
            f"{('ORDER BY ' + ', '.join(orders)) if orders else ''} "
            f"LIMIT {limit} OFFSET {offset}"
        )

        return list(map(self.to_entity_type, self.execute(sql)[1]))

    def find_by_id(self, entity_id: Identity, columns: Optional[Set[str]] = None) -> Optional[E]:
        columns = "*" if not columns else ", ".join(columns)
        clauses = [f"{k} = {quote(v)}" for k, v in zip(entity_id.attributes, entity_id.values)]
        sql = f"SELECT {columns} FROM {self.table_name()} " f"WHERE {' AND '.join(clauses)}"
        result = next((e for e in self.execute(sql)[1]), None)

        return self.to_entity_type(result) if result else None

    def exists_by_id(self, entity_id: Identity) -> bool:
        clauses = [f"{k} = {quote(v)}" for k, v in zip(entity_id.attributes, entity_id.values)]
        sql = f"SELECT EXISTS(SELECT 1 FROM {self.table_name()} WHERE {' AND '.join(clauses)})"

        return self.execute(sql)[1][0][0] > 0

    @abstractmethod
    def table_name(self) -> str:
        pass

    @abstractmethod
    def to_entity_type(self, entity_dict: dict | tuple) -> E:
        pass



# _*_ coding: utf-8 _*_
#
# hspylib-datasource v0.9.40
#
# Package: main.datasource.redis
"""Package initialization."""

__all__ = [
    'redis_configuration', 
    'redis_repository'
]
__version__ = '0.9.40'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib-Datasource
   @package: datasource
      @file: crud_repository.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from abc import abstractmethod
from datasource.crud_entity import CrudEntity
from datasource.identity import Identity
from hspylib.core.metaclass.singleton import AbstractSingleton
from hspylib.core.namespace import Namespace
from typing import Generic, List, Optional, Set, TypeVar

import json

E = TypeVar("E", bound=CrudEntity)


class CrudRepository(Generic[E], metaclass=AbstractSingleton):
    """Interface for generic CRUD operations on a repository for a specific type."""

    @property
    def logname(self) -> str:
        return self.__class__.__name__.split("_", maxsplit=1)[0]

    @abstractmethod
    def count(self) -> int:
        """Returns the number of entities available."""

    @abstractmethod
    def delete(self, entity: E) -> None:
        """Deletes a given entity.
        :param entity: the entity to delete.
        """

    @abstractmethod
    def delete_by_id(self, entity_id: Identity) -> None:
        """Deletes the entity with the given id.
        :param entity_id: the ID of the entity to be deleted.
        """

    @abstractmethod
    def delete_all(self, entities: List[E]) -> None:
        """Deletes all given entities.
        :param entities: the entities to be deleted.
        """

    @abstractmethod
    def save(self, entity: E) -> None:
        """Saves a given entity.
        :param entity: the entity to save.
        """

    @abstractmethod
    def save_all(self, entities: List[E]) -> None:
        """Saves all given entities.
        :param entities: the entities to be saved.
        """

    @abstractmethod
    def find_all(
        self,
        columns: Set[str] | None = None,
        filters: Namespace | None = None,
        order_bys: Set[str] | None = None,
        limit: int = 500,
        offset: int = 0,
    ) -> List[E]:
        """Returns all entities of the type.
        :param columns: the column names to select.
        :param filters: entry filters.
        :param order_bys: result set order bys.
        :param limit: the maximum number of entries to be fetch.
        :param offset: skip offset entries before fetch.
        """

    @abstractmethod
    def find_by_id(self, entity_id: Identity, columns: Set[str] | None = None) -> Optional[E]:
        """Retrieves an entity by its id.
        :param entity_id: the entity ID.
        :param columns: the column names to select.
        """

    def exists_by_id(self, entity_id: Identity) -> bool:
        """Returns whether an entity with the given id exists.
        :param entity_id: the entity ID.
        """

    @abstractmethod
    def to_entity_type(self, entity_dict: dict | tuple) -> E:
        """Convert a dict or tuple, generally from a result set, into the CRUD entity.
        :param entity_dict: the entity dict or tuple to be converted.
        """

    def to_entity_list(self, json_string: str, filters: Namespace | None = None) -> List[E]:
        """Return filtered entries from the json_string as a list.
        :param json_string: the json string to be parsed.
        :param filters: entry filters.
        """
        ret_list = []
        for value in json.loads(json_string).values():
            if not filters or all(k in value and value[k] == v for k, v in filters.items()):
                ret_list.append(self.to_entity_type(value))
        return ret_list



# _*_ coding: utf-8 _*_
#
# hspylib-datasource v0.9.40
#
# Package: main.datasource
"""Package initialization."""

__all__ = [
    'crud_entity', 
    'crud_repository', 
    'crud_service', 
    'db_configuration', 
    'db_repository', 
    'exception', 
    'firebase', 
    'identity', 
    'mysql', 
    'redis', 
    'sqlite'
]
__version__ = '0.9.40'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib-Datasource
   @package: datasource.exception
      @file: exceptions.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""

from hspylib.core.exception.exceptions import HSBaseException


class EntityNotFoundError(HSBaseException):
    """Raised when an entity is not found"""


class DatabaseConnectionError(HSBaseException):
    """Raised when all connection attempts to the database exhausted"""


class DatabaseError(HSBaseException):
    """Raised when database failed to execute/create-session"""


class ProgrammingError(HSBaseException):
    """Exception raised for programming errors, e.g. table not found
    or already exists, syntax error in the SQL statement, wrong number
    of parameters specified, etc."""
